//
// $Id: dns.cc,v 1.2 1999/08/06 08:44:33 voeckler Exp $
//
// Author:  Jens-S. Vckler <voeckler@rvs.uni-hannover.de>
//
// File:    dns.cc
//          Thu Aug 05 1999
//
// (c) 1999 Lehrgebiet Rechnernetze und Verteilte Systeme
//          Universitt Hannover, Germany
//
// Permission to use, copy, modify, distribute, and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that (i) the above copyright notices and this permission
// notice appear in all copies of the software and related documentation,
// and (ii) the names of the Lehrgebiet Rechnernetze und Verteilte
// Systeme and the University of Hannover may not be used in any
// advertising or publicity relating to the software without the
// specific, prior written permission of Lehrgebiet Rechnernetze und
// Verteilte Systeme and the University of Hannover.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL THE LEHRGEBIET RECHNERNETZE UND VERTEILTE SYSTEME OR
// THE UNIVERSITY OF HANNOVER BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
// INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
// SOFTWARE.
//
// $Log: dns.cc,v $
// Revision 1.2  1999/08/06 08:44:33  voeckler
// added debugging to DNS using the debug level and debug class from global.
//
// Revision 1.1  1999/08/05 21:06:01  voeckler
// Initial revision
//
//
#if defined(__GNUC__) || defined(__GNUG__)
#pragma implementation
#endif

#include <assert.h>
#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <fcntl.h>
#include <string.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "global.hh"
#include "ctype.hh"
#include "dns.hh"

static const char* RCS_ID =
"$Id: dns.cc,v 1.2 1999/08/06 08:44:33 voeckler Exp $";

DNSCache::DNSCache( const char* cacheFile,
		    MyUInt32 posTTL,
		    MyUInt32 negTTL )
  :BaseCache(posTTL,negTTL),db(0)
{
  assert(cacheFile != 0);
  if ( (db = dbm_open( (char*) cacheFile, O_RDWR | O_CREAT, 0664 )) == 0 ) {
    fprintf( stderr, "dbm_open(%s): unable to open DB file!\n", cacheFile );
    exit(1);
  }
}

DNSCache::~DNSCache()
{
  if ( dbm_error(db) ) dbm_clearerr(db);
  dbm_close(db);
}

static
MyUInt32
_retrieve( DBM* db, const char* key, char* result, size_t rsize )
{
  size_t size = strlen(key)+1;
  char* local = new char[size];
  for ( int i=0; i<size; ++i ) local[i] = TOLOWER(key[i]);

  // terminate result
  result[0] = result[--rsize] = '\0';

  datum _key;
  _key.dptr = local;
  _key.dsize = strlen(key)+1;

  datum _value = dbm_fetch( db, _key );
  if ( dbm_error(db) ) dbm_clearerr(db);
  if ( (globals.debugClass & Global::D_DNS) ) {
    fprintf( stderr, "# DNS: dbm_fetch(%s): %lu byte @ %p\n", 
	     local, _value.dsize, _value.dptr );
  }
  delete[] local;
  if ( _value.dptr == 0 ) return C_U32(0); // not found

  MyUInt32 ttl;
  memcpy( &ttl, _value.dptr, sizeof(MyUInt32) );
  strncpy( result, ((char*) _value.dptr)+sizeof(MyUInt32), rsize );
#if USES_GDBM
  free((void*) _value.dptr );
#endif
  return ttl;
}

void
DNSCache::_enter( struct hostent* h, const char* key, MyUInt32 now )
{
  struct in_addr ip;
  datum _key, _value;
  
  if ( h && strstr( h->h_name, "in-addr.arpa" ) == 0 ) {
    // positive lookup
    if ( (globals.debugClass & Global::D_DNS) ) {
      fprintf( stderr, "# DNS positive lookup: %s", h->h_name );
      for ( char** p=h->h_aliases; *p; ++p )
	fprintf( stderr, " %s", *p );
      for ( char** p=h->h_addr_list; *p; ++p ) {
	ip.s_addr = *((MyUInt32*) *p);
	fprintf( stderr, " %s", inet_ntoa(ip) );
      }
      fputc( '\n', stderr );
    }

    now += positiveTTL;
    memcpy( &ip, h->h_addr_list[0], sizeof(ip) );
    size_t size = sizeof(now) + 17;
    char* local = new char[size];
    memcpy( local, &now, sizeof(now) );
    strncpy( local+sizeof(now), inet_ntoa(ip), 16 );
    _value.dptr = local;
    _value.dsize = size;

    for ( char** p=h->h_aliases; *p; ++p )
      if ( strstr( *p, "in-addr.arpa" ) == 0 ) {
	size = strlen(*p)+1;
	char* what = new char[size];
	for ( int i=0; i<size; ++i ) what[i] = TOLOWER((*p)[i]);
	_key.dptr = what;
	_key.dsize = size;
	if ( (globals.debugClass & Global::D_DNS) && globals.debugLevel )
	  fprintf( stderr, "# DNS storing %s --> %s\n", what, local+4 );
	if ( dbm_store( db, _key, _value, DBM_REPLACE ) != 0 )
	  fprintf( stderr, "# DNS store(%s): error %d\n", what, dbm_error(db));
	delete[] what;
      }

    size = strlen(h->h_name)+1;
    char* what = new char[size];
    for ( int i=0; i<size; ++i ) what[i] = TOLOWER(h->h_name[i]);
    _key.dptr = what;
    _key.dsize = size;
    if ( (globals.debugClass & Global::D_DNS) && globals.debugLevel )
      fprintf( stderr, "# DNS storing %s --> %s\n", what, local+4 );
    if ( dbm_store( db, _key, _value, DBM_REPLACE ) != 0 )
      fprintf( stderr, "# DNS store(%s): error %d\n", what, dbm_error(db) );

    delete[] local;
    size += sizeof(now);
    local = new char[size];
    memcpy( local, &now, sizeof(now) );
    strncpy( local+sizeof(now), what, size-sizeof(now) );
    delete[] what;
    _value.dptr = local;
    _value.dsize = size;

    for ( char** p=h->h_addr_list; *p; ++p ) {
      ip.s_addr = *((MyUInt32*) *p);
      size = strlen(inet_ntoa(ip))+1;
      what = new char[size];
      strncpy( what, inet_ntoa(ip), size );
      _key.dptr = what;
      _key.dsize = size;
      if ( (globals.debugClass & Global::D_DNS) && globals.debugLevel )
	fprintf( stderr, "# DNS storing %s --> %s\n", what, local+4 );
      if ( dbm_store( db, _key, _value, DBM_REPLACE ) != 0 )
	fprintf( stderr, "# DNS store(%s): error %d\n", what, dbm_error(db) );
      delete[] what;
    }
    delete[] local;

  } else {
    // negative lookup
    if ( (globals.debugClass & Global::D_DNS) ) {
      fprintf( stderr, "# DNS negative lookup: %s\n", key );
    }

    now += negativeTTL;
    size_t size = sizeof(now)+1;
    char* local = new char[size];
    memcpy( local, &now, sizeof(now) );
    local[sizeof(now)] = '\0';
    _value.dptr = local;
    _value.dsize = size;

    size = strlen(key)+1;
    char* what = new char[size];
    for ( int i=0; i<size; ++i ) what[i] = TOLOWER(key[i]);
    _key.dptr = what;
    _key.dsize = size;
    dbm_store( db, _key, _value, DBM_REPLACE );
    delete[] what;
    delete[] local;
  }
}

size_t
DNSCache::gethostbyname( const char* what, char* result, size_t rsize )
{
  MyUInt32 now = time(0);
  requests[C_ALL]++;
  MyUInt32 ttl = _retrieve( db, what, result, rsize );
  if ( ttl < now ) {
    // not found or expired
    _enter( ::gethostbyname(what), what, now );
    ttl = _retrieve( db, what, result, rsize );
  } else {
    if ( (globals.debugClass & Global::D_DNS) && globals.debugLevel )
      fprintf( stderr, "# DNS hit: %s --> %s\n", what, result );
    requests[C_HIT]++;
  }
  return strlen(result);
}

size_t
DNSCache::gethostbyaddr( const char* what, char* result, size_t rsize )
{
  MyUInt32 addr, now = time(0);
  requests[C_ALL]++;
  MyUInt32 ttl = _retrieve( db, what, result, rsize );
  if ( ttl < now ) {
    // not found or expired
    if ( (addr = inet_addr(what)) == C_U32(-1) ) {
      fprintf( stderr, "# DNS: illegal host address \"%s\"\n", what );
      strncpy( result, what, rsize-1 );
      result[rsize-1] = '\0';
      return 0;
    }
    struct hostent* h = ::gethostbyaddr( (const char*) &addr,
					 sizeof(addr), AF_INET );
    _enter( h, what, now );
    ttl = _retrieve( db, what, result, rsize );
  } else {
    if ( (globals.debugClass & Global::D_DNS) && globals.debugLevel )
      fprintf( stderr, "# DNS hit: %s --> %s\n", what, result );
    requests[C_HIT]++;
  }
  return strlen(result);
}


