//
// $Id: seafood.cc,v 1.4 1999/08/08 09:07:42 voeckler Exp $
//
// Author:  Jens-S. Vckler <voeckler@rvs.uni-hannover.de>
//
// File:    newcm.hh
//          Thu Aug 05 1999
//
// (c) 1999 Lehrgebiet Rechnernetze und Verteilte Systeme
//          Universitt Hannover, Germany
//
// Permission to use, copy, modify, distribute, and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that (i) the above copyright notices and this permission
// notice appear in all copies of the software and related documentation,
// and (ii) the names of the Lehrgebiet Rechnernetze und Verteilte
// Systeme and the University of Hannover may not be used in any
// advertising or publicity relating to the software without the
// specific, prior written permission of Lehrgebiet Rechnernetze und
// Verteilte Systeme and the University of Hannover.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL THE LEHRGEBIET RECHNERNETZE UND VERTEILTE SYSTEME OR
// THE UNIVERSITY OF HANNOVER BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
// INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
// SOFTWARE.
//
// $Log: seafood.cc,v $
// Revision 1.4  1999/08/08 09:07:42  voeckler
// added compile-time configurable libz and libbz2 support.
//
// Revision 1.3  1999/08/06 13:07:08  voeckler
// added stats to fsm01 in order to parse "CONNECT some.host:443" correctly.
//
// Revision 1.2  1999/08/05 22:09:30  voeckler
// made the config file location commandline configurable, added a force
// no lookups flag.
//
// Revision 1.1  1999/08/05 21:16:27  voeckler
// Initial revision
//
//
#if defined(__GNUC__) || defined(__GNUG__)
#pragma implementation
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>

#include <sys/types.h>
#include <sys/time.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>

#include "typedefs.h"
#include "global.hh"
#include "counter.hh"
#include "ctype.hh"
#include "hash.hh"
#include "string.hh"
#include "input.hh"
#include "logline.hh"
#include "counters.hh"
#include "tools.hh"
#include "fsm01.hh"
#include "output.hh"
#include "dns.hh"

static const char*  RCS_ID =
"$Id: seafood.cc,v 1.4 1999/08/08 09:07:42 voeckler Exp $";

// global var to communication with flex/bison
int parserErrorSeen;
extern int yyparse();

BaseInput*
obtainFile( const char* fn ) 
  // purpose: check the given file name for being of the 'right' kind
  //          and return either a gunzip or a plain input stream to it.
  // paramtr: fn (IN): file name
  // returns: opened input stream, of NULL in case of error
{
  BaseInput* result = 0;

  // try to open the specified file
  int fd = open( fn, O_RDONLY | O_NOCTTY );
  if ( fd == -1 ) {
    fprintf( stderr, "open(%s): %s\n", fn, strerror(errno) );
    return 0;
  }
  
  // check for gzipped input by reading a few byte (see: /etc/magic)
  unsigned char test[4];
  int option;
  do {
    option = read( fd, test, 4 );
  } while ( option == -1 && errno == EINTR );

  // see how much we have got, complain bitterly otherwise
  if ( option < 0 ) {
    perror( "read from input" );
    close(fd);
    return 0;
  } else if ( option < 4 ) {
    fputs( "read from input: premature end of input\n", stderr );
    close(fd);
    return 0;
  }

  // mind, we cannot seek pipes nor sockets
  if ( lseek( fd, -4, SEEK_CUR ) == -1 ) {
    perror( "unable to seek" );
    close(fd);
    return 0;
  }

  if ( ISDIGIT(test[0]) && ISDIGIT(test[1]) && ISDIGIT(test[2]) ) {
    // plain input
    result = new PlainInput( fd, 65520 );
  }
#ifdef USE_LIBZ
    else if ( (test[0] == 0x1F) && (test[1] == 0x8B) && (test[2] <= 0x08) ) {
    // gzip input
    result = new GZipInput( fd, 65520 );
  }
#endif // USE_LIBZ
#ifdef USE_LIBBZ2
    else if ( (test[0] == 0x42) && (test[1] == 0x5a) && (test[2] == 0x68) &&
	      (test[3] >= 0x30) ) {
    // bzip2 input
    result = new BZip2Input( fd, 65520 );
  }
#endif // USE_LIBBZ2
    else {
    // unknown input format
    fputs( "unknown input format (neither gzip, bzip2 nor plain)\n", stderr );
  }

  return result;
}

double
processFile( BaseInput* in, Counters& cnt )
  // purpose: big loop around file for parsing the contents
  // paramtr: in (IO): input stream to read until it ends
  // returns: time spent processing this file
{
  char temp[8192];
  FSM_01 fsm;
  LogLine log;
  int urlsize, column;
  bool runOnce = true;
  String log_tld, log_sld, log_hierhost, log_client;

  MyUInt32 lastStamp = 0;	// copy of last timestamp
  unsigned long linenoCopy;	// file based line nr after LF was read

  char* const urlComponents[3] = { temp, temp+256, temp+(256+256) };
  const size_t urlCompSizes[3] = { 256, 256, 32 };

  // fixes for Squid false log messages
#ifdef REFRESH_FIX
  int tcp_refresh_hit = LogLine::stable.find("TCP_REFRESH_HIT") & 0xFFFF;
#endif // REFRESH_FIX
  int hier_none = LogLine::htable.find("NONE") | IS_NONE;

  double start = now();
  while ( ! in->eof() ) {
    // report the line we are currently looking at.
    if ( (in->lineno & 65535) == 0 )
      fprintf( stderr, "# %lu lines processed from this file\n", in->lineno );

    //
    // STAGE 1: start parsing a log line
    //
    log.shallow_clear();
    column = 0;

    // 1st column, timestamp, e.g. 931212014.216
    // just read the integer portion and discard fractional value
    if ( in->integer( log._stamp ) < 9 ) {
      fprintf( stderr, "%lu: illegal timestamp, skipping\n", in->lineno );
    } else {
      in->skiptolws();
      column++;
    }

    // 2nd column, duration, e.g. 25812, or 0, or even -1
    if ( in->integer( log._duration ) == 0 ) {
      fprintf( stderr, "%lu: illegal duration, skipping\n", in->lineno );
    } else
      column++;


    // 3rd column, client address, e.g. "130.75.3.218" or "rot-lane.rvs...."
    if ( in->get( log._requester, sizeof(log._requester) ) < 3 ) {
      fprintf( stderr, "%lu: illegal client host address, skipping\n", 
	       in->lineno );
    } else
      column++;


    // 4th column, cache hit status, e.g.: "TCP_MISS/200"
    if ( log.parseStatus( in ) < 6 || log._statuscode == -1 ) {
      fprintf( stderr, "%lu: illegal return status, skipping\n", in->lineno );
    } else
      column++;


    // 5th column, reply size, e.g. 8838
    if ( in->integer( log._size ) == 0 ) {
      fprintf( stderr, "%lu: illegal reply size, skipping\n", in->lineno );
    } else
      column++;


    // 6th column, method string, e.g. "GET"
    if ( log.parseMethod( in ) < 3 || log._method == 0 ) {
      fprintf( stderr, "%lu: unknown method, skipping\n", in->lineno );
    } else
      column++;

    // 7th column, URI, e.g. "http://www.user.cityline.ru:8083/~itbshake/"
    if ( (urlsize = in->get( log._url, sizeof(log._url)) ) < 5 ) {
      fprintf( stderr, "%lu: strange URL, skipping\n", in->lineno );
    } else
      column++;
    
    // 8th column, ident, e.g. "-"
    if ( globals.noIdent ) {
      // precondition: no ident lookups, value will always be "-"
      // if it isn't, the URL contains whitespaces (which are illegal)
      // FIXME: what happens, if the line end prematurely ?!
      while ( in->get( temp, sizeof(temp) ) != 1 && *temp != '-' ) {
#if 0 // NOT YET NEEDED, SAFE TIME HERE!
	// whoa! there were whitespaces in the url...
	strncat( log._url, "%20", sizeof(log._url) );
	strncat( log._url, temp, sizeof(log._url) );
	urlsize += strlen(temp) + 3;
#endif
	fprintf( stderr, "%lu: URL contains whitespaces, fixing!\n", 
		 in->lineno );
      }

      // we found what we were looking for (hopefully).
      log._ident[0] = '-';
      log._ident[1] = '\0';
      column++;
    } else {
      // precondition: ident lookups --> url ws cognition system is down :-(
      if ( in->get( log._ident, sizeof(log._ident) ) == 0 ) {
	fprintf( stderr, "%lu: unknown ident string, skipping\n", in->lineno );
      } else
        column++;
    }

    // 9th column, hierarchy info, e.g. "DIRECT/www.user.cityline.ru"
    if ( log.parseHierarchy(in) == 0 || log._hierarchy == 0 ) {
      fprintf( stderr, "%lu: unknown hierarchy, skipping\n", in->lineno );
    } else
      column++;

    // 10th column, mime type, e.g. "text/html"
    // Illegal media types are 'valid' and counted as error.
    // Only parse media type for TCP et. al., skip for UDP
    if ( (log._status & IS_UDP) == 0 ) {
      if ( log.parseMediaType(in) == 0 ) {
	fprintf( stderr, "%lu: strange MIME type, skipping\n", in->lineno );
      } else
	column++;
    } else {
      column++;
    }

    // 11th and 12th column may contain request and reply headers, skip
    // also, skippasteoln() will update the line counter and positioner
    linenoCopy = in->lineno;
    in->skippasteoln();
    if ( column < 10 ) {
      log.dump( stderr, 10 );
      continue;
    }

    //
    // STAGE 2: convert some values into something even more computable
    //
    unsigned short log_port = 0;
    int log_scheme = 0;
    log_client = log._requester;

    if ( runOnce ) {
      if ( globals.startTime == 0 || globals.stopTime > log._stamp )
	globals.startTime = log._stamp;
      runOnce = false;
    }
    if ( globals.stopTime < log._stamp ) 
      globals.stopTime = log._stamp;
    MyUInt32 log_peak = log._stamp / globals.peakInterval;

    if ( lastStamp && 
	 abs( log._stamp - lastStamp ) > globals.warnCrashInterval ) {
      char oldstamp[32], newstamp[32];
      time_t t1 = lastStamp;
      time_t t2 = log._stamp;
      strftime( oldstamp, 32, "%d.%m.%Y %H:%M:%S %Z", localtime(&t1) );
      strftime( newstamp, 32, "%d.%m.%Y %H:%M:%S %Z", localtime(&t2) );
      fprintf( stderr, "%lu: detected %.0f s silence between two time stamps\n"
	       "+ last \"%s\", current \"%s\"\n", linenoCopy,
	       difftime(t2,t1), oldstamp, newstamp  );
    }
    lastStamp = log._stamp;

    if ( (log._status & IS_UDP) == 0 ) {
      // determine scheme, hostname (tld,2ld) and portnumber
      fsm.parse( log._url, urlComponents, urlCompSizes );
      if ( (log_scheme = LogLine::schemes.find(urlComponents[0])) == 0 &&
	   globals.warnUnknownScheme )
	fprintf( stderr, "%lu: unknown URL scheme \"%s\"\n", 
		 linenoCopy, urlComponents[0] );
      getDomainFromHost( urlComponents[1], log_tld, log_sld );
      if ( log_tld.length() == 0 || LogLine::domains.find(log_tld()) == 0 ) {
	if ( globals.warnUnknownTLD )
	  fprintf( stderr, "%lu: unknown TLD \"%s\" in hostname \"%s\"\n",
		   linenoCopy, log_tld(), urlComponents[1] );
	log_tld = str_error;
	log_sld = str_error;
      }

      log_port =
	ISDIGIT( urlComponents[2][0] ) ?
	strtol(urlComponents[2],0,10) & 0xFFFF :
	80; // FIXME: check with getservbyname()
      log_port = (( log_port == 80 ) ? 64 : ( log_port >> 10 ));
      log_hierhost = String( log._hierhost );
    }
    
    // use 'natural' sizes for indexing, should be fastest
    int log_status  = log._status & 0xFFFF;
    int log_method  = log._method & 0xFFFF;
    int log_media   = log._mediatype & 0xFFFF;
    int log_subtype = log._mediasubtype & 0xFFFF;

    // fix Squid false log messages
#define TCP_HIT (IS_HIT | IS_TCP)
#ifdef REFRESH_FIX
    if ( log_status == tcp_refresh_hit && log._statuscode == 200 )
      log._status &= ~IS_HIT;
    else
#endif // REFRESH_FIX
    if ( (log._status & TCP_HIT) == TCP_HIT &&
	 ( log._statuscode == 301 || log._statuscode == 302 ) )
      log._hierarchy = hier_none;

    //
    // STAGE 3: count!
    //
#ifdef LOOP_ADD
    cnt.all.incr( log._size, log._duration );
#endif
    cnt.methods[log_method ].incr( log._size, log._duration );

    if ( (log._status & IS_UDP) != 0 ) {
      // count all UDP traffic here
#ifdef LOOP_ADD
      cnt.udp.incr( log._size, log._duration );
#endif
      cnt.udp_client[ log_client ].incr( log._size, log._duration );
      cnt.peak_udp[ log_peak ].incr( log._size, log._duration );
      if ( (log._status & IS_HIT) != 0 ) {
	// count UDP HITs here
	cnt.udp_hit.incr( log._size, log._duration );
	cnt.udp_hit_client[ log_client ].incr( log._size, log._duration );
	cnt.udp_hit_status[ log_status ].incr( log._size, log._duration );
	cnt.hit_methods[ log_method ].incr( log._size, log._duration );
	cnt.peak_udp_hit[ log_peak ].incr( log._size, log._duration );
      } else {
	// count UDP MISSes here
	cnt.udp_miss.incr( log._size, log._duration );
	cnt.udp_miss_status[ log_status ].incr( log._size, log._duration );
      }

    } else {
      if ( (log._status & IS_TCP) == 0 ) {
	fprintf( stderr, "%lu: neither TCP nor UDP\n", linenoCopy );
	log.dump(stderr,10);
      }

      int log_hier = log._hierarchy & 0xFFFF;

      // count all ***OTHER*** traffic here (not limited to TCP)
#ifdef LOOP_ADD
      cnt.tcp.incr( log._size, log._duration );
      cnt.tcp_client[ log_client ].incr( log._size, log._duration );
#endif
      cnt.tcp_tld[ log_tld ].incr( log._size, log._duration );
      cnt.tcp_sld[ log_sld ].incr( log._size, log._duration );
      cnt.tcp_scheme[ log_scheme ].incr( log._size, log._duration );
      cnt.tcp_mime[ log_media ].incr( log._size, log._duration );
      if ( cnt.nSubtype[log_media] )
	cnt.tcp_submime[ log_media ][ log_subtype ].incr( log._size, 
							  log._duration );
      cnt.tcp_ports[ log_port ].incr( log._size, log._duration );
      cnt.peak_tcp[ log_peak ].incr( log._size, log._duration );

      if ( (log._status & TCP_HIT) == TCP_HIT ) {
	// TCP HITs
	cnt.tcp_hit.incr( log._size, log._duration );
	cnt.tcp_hit_client[ log_client ].incr( log._size, log._duration );
	cnt.tcp_hit_status[ log_status ].incr( log._size, log._duration );
	cnt.tcp_hit_tld[ log_tld ].incr( log._size, log._duration );
	cnt.tcp_hit_sld[ log_sld ].incr( log._size, log._duration );
	cnt.tcp_hit_mime[ log_media ].incr( log._size, log._duration );
	if ( cnt.nSubtype[log_media] )
	  cnt.tcp_hit_submime[ log_media ][ log_subtype ].incr( log._size, 
								log._duration );
	cnt.tcp_hit_scheme[ log_scheme ].incr( log._size, log._duration );
	cnt.hit_methods[ log_method ].incr( log._size, log._duration );

	cnt.tcp_hit_hier[ log_hier ].incr( log._size, log._duration );
	if ( (log._hierarchy & (IS_DIRECT|IS_NONE))==IS_DIRECT && log_hier )
	  cnt.tcp_hit_direct_host[ log_hierhost ].incr( log._size,
							log._duration );
	cnt.peak_tcp_hit[ log_peak ].incr( log._size, log._duration );
      } else if ( (log._hierarchy & IS_NONE) || log._hierarchy == 0 ) {
	// MISSes with NONE (mostly errors)
	cnt.tcp_miss_none.incr( log._size, log._duration );
	cnt.tcp_miss_none_status[ log_status ].incr( log._size, log._duration );
	cnt.tcp_miss_none_client[ log_client ].incr( log._size, log._duration );
      } else {
	// TCP MISSes
	cnt.tcp_miss.incr( log._size, log._duration );
	cnt.tcp_miss_status[ log_status ].incr( log._size, log._duration );
	cnt.tcp_miss_client[ log_client ].incr( log._size, log._duration );
      }

      //
      // for all ***OTHER*** traffic, look into the hierarchy
      //
      if ( log_hier && (log._hierarchy & IS_NONE) == 0 ) {
	// any kind of traffic with attached hierarchy code
#ifdef LOOP_ADD
	cnt.hier.incr( log._size, log._duration );
#endif
	//---cnt.hier_method[ log_hier ].incr( log._size, log._duration );
	// cnt.hier_host[ log_hierhost ].incr( log._size, log._duration );
	// cnt.hier_status[ log_hier ][ log_hierhost ].incr( log._size, 
	//						     log._duration );
	if ( (log._hierarchy & IS_DIRECT) != 0 ) {
	  // directly to destination transcentral
	  cnt.hier_direct.incr( log._size, log._duration );
	  cnt.hier_direct_method[ log_hier ].incr( log._size, log._duration );
	  cnt.hier_direct_host[ log_hierhost ].incr( log._size, log._duration );
	  cnt.peak_hier_direct[ log_peak ].incr( log._size, log._duration );
	} else if ( (log._hierarchy & IS_PEER) != 0 ) {
	  // any peer business (carp, cache digests, etc)
	  cnt.hier_peer.incr( log._size, log._duration );
	  cnt.hier_peer_method[ log_hier ].incr( log._size, log._duration );
	  cnt.hier_peer_host[ log_hierhost ].incr( log._size, log._duration );
	  cnt.hier_peer_status[ log_hier ][ log_hierhost ].incr( log._size, 
		log._duration );
	  cnt.peak_hier_peer[ log_peak ].incr( log._size, log._duration );
	} else if ( (log._hierarchy & IS_PARENT) != 0 ) {
	  cnt.hier_parent.incr( log._size, log._duration );
	  cnt.hier_parent_method[ log_hier ].incr( log._size, log._duration );
	  cnt.hier_peer_host[ log_hierhost ].incr( log._size, log._duration );
	  cnt.hier_peer_status[ log_hier ][ log_hierhost ].incr( log._size, 
		log._duration );
	  cnt.peak_hier_parent[ log_peak ].incr( log._size, log._duration );
	} else {
	  fprintf( stderr, "%lu: unknown hierarchy!\n", linenoCopy );
	}
      }	// hierarchy traffic
    } // !UDP
  } // big loop

#ifndef LOOP_ADD
  // save cycles by *not* counting deducables values in the loop
  cnt.hier = cnt.hier_direct + cnt.hier_peer + cnt.hier_parent;
  cnt.tcp = cnt.tcp_hit + cnt.tcp_miss_none + cnt.tcp_miss;
  cnt.udp = cnt.udp_hit + cnt.udp_miss;
  cnt.all = cnt.udp + cnt.tcp;

  for ( CountMapIterator i(cnt.tcp_hit_client); i.avail(); i++ )
    cnt.tcp_client[i.key()] += i();
  for ( CountMapIterator i(cnt.tcp_miss_client); i.avail(); i++ )
    cnt.tcp_client[i.key()] += i();
  for ( CountMapIterator i(cnt.tcp_miss_none_client); i.avail(); i++ )
    cnt.tcp_client[i.key()] += i();
#endif

  fprintf( stderr, "# %lu lines processed from this file\n", in->lineno-1 );
  return now() - start;
}

void
readConfigFile( const char* fn )
{
  extern FILE* yyin;

  fprintf( stderr, "# trying to read \"%s\"\n", fn );
  if ( (yyin=fopen(fn,"r")) ) {
    yyparse();
    fclose(yyin);
    if ( parserErrorSeen ) {
      fprintf( stderr, "there were errors in the config file, exiting!\n" );
      exit(1);
    }
  } else {
    fprintf( stderr, "unable to read config file \"%s\": %s\n", 
	     fn, strerror(errno) );
  }
  fprintf( stderr, "# done reading \"%s\"\n", fn );
}

static
void
helpMe( const char* argv0, const char* rcsid )
{
  // obtain my own basename...
  char* s = strrchr(argv0,'/');
  if ( s ) ++s;

  printf( "%s\nUsage:\t%s [-f cfgfile] file1 [file2 [..]]\t"
          "# plain, gzip, or bzip2\n", rcsid, s );
  printf( " -f fn\tUse configuration file fn instead of %s.conf\n", argv0 );
  puts( " -N\tDisable all DNS and IRR lookups regardless of configuration "
	"content" );
  puts( "" );
  exit(1);
}

static
void
parseCommandline( int argc, char* argv[], 
		  char*& configFilename,
		  bool& forceDisableLookups )
  // purpose: parse optional arguments from the command line
  // paramtr: argc (IN): see main()
  //          argv (IN): see main()
  //          configFilename (OUT): place of configuration file
  //          forceDisableLooups (OUT): true==disable DNS+IRR 
{
  // construct configuration file default name
  configFilename = new char[strlen(argv[0])+6];
  strcpy( configFilename, argv[0] );
  strcat( configFilename, ".conf" );
  forceDisableLookups = false;

  // parse options
  int option;
  while ( (option = getopt( argc, argv, "Nf:" )) != -1 ) {
    switch ( option ) {
    case 'f':
      delete[] configFilename;
      configFilename = new char[strlen(optarg)+1];
      strcpy( configFilename, optarg );
      break;
    case 'N':
      forceDisableLookups = true;
      break;
    default:
      helpMe(argv[0],RCS_ID);
      break;
    }
  }
}

int
main( int argc, char* argv[] )
{
  double startup = now();
  double looptime = 0.0;
  unsigned long lineno = 0;

  // read some basic configuration commands
  bool disableLookups = false;
  char* configuration = 0;
  parseCommandline( argc, argv, configuration, disableLookups );
  readConfigFile( configuration );
  delete[] configuration;

  // force disable lookups, if such was commandline-wanted
  // that will still open the DNS database...
  if ( disableLookups ) {
    globals.clientFQDN = false;
    if ( globals.irrServer ) free((void*) globals.irrServer);
    globals.irrServer = 0;
  }

  // in case of debugging, use slower line buffered output
  if ( globals.debugClass || globals.debugLevel )
    setvbuf( stdout, 0, _IOLBF, 8192 );

  // *AFTER* reading configuration, allocate counters
  Counters counters; // default c'tor
  IRRCache* irr = 0; // IRR connection
  DNSCache dns( globals.dnsCacheFile, 
		globals.dnsPositiveTTL, globals.dnsNegativeTTL );
  if ( globals.irrServer ) 
    irr = new IRRCache( globals.irrCacheFile, globals.irrServer,
			globals.irrPositiveTTL, globals.irrNegativeTTL );

  for ( int filenr=optind; filenr < argc; filenr++ ) {
    BaseInput* in = obtainFile( argv[filenr] );
    if ( in ) { 
      looptime += processFile( in, counters );
      lineno += (in->lineno - 1);
    }
    delete in;
  }

  fputs( "# done processing files, writing results\n", stderr );
  if ( looptime < 1E-4 ) looptime = 1E-4;
  fputs( "# title...\n", stderr );
  showTitle();
  fflush(stdout);

  // overview
  fputs( "# overview...\n", stderr );
  showOverview(counters);
  fflush(stdout);

  // hierarchy info
  fputs( "# hierarchy...\n", stderr );
  showHierarchy(dns,counters);
  fflush(stdout);

  // methods
  fputs( "# methods...\n", stderr );
  showMethods(counters);
  fflush(stdout);

  // schemes
  fputs( "# schemes...\n", stderr );
  showSchemes(counters);
  fflush(stdout);

  // top-level domains
  fputs( "# tlds...\n", stderr );
  showTLD( counters, 0 );
  fflush(stdout);

  // 2nd-level domains
  fputs( "# 2lds...\n", stderr );
  showSLD( counters, 20 );
  fflush(stdout);

  if ( irr ) {
    // show directs
    fputs( "# asn...\n", stderr );
    showASN( dns, *irr, counters, 20 );
    fflush(stdout);
  }

  // media types
  fputs( "# mime...\n", stderr );
  showMediatypes(counters);
  fflush(stdout);

  // show port statistics
  fputs( "# ports...\n", stderr );
  showPorts( counters );
  fflush(stdout);

  // show clients
  fputs( "# clients...\n", stderr );
  showUDPClients( dns, counters, 20 );
  fflush(stdout);
  showTCPClients( dns, counters, 20 );
  fflush(stdout);

  // show peak intervals
  fputs( "# peaks...\n", stderr );
  showPeaks( counters );
  fflush(stdout);
  
  // statistics about the log file analysis
  fputs( "# stats...\n", stderr );
  showStats( dns, irr, looptime, startup, lineno );
  fflush(stdout);

  fputs( "# DONE!\n", stderr );
  delete irr;
  return 0;
}
